﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Flash: Zugriff auf Festspeicher            											//
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2022 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "stm32f3xx.h"
#include "Flash.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
WaitForIdle(void)
{
	while ((bool)READ_BIT(FLASH->SR, FLASH_SR_BSY));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
Unlock(void)
{
	WaitForIdle();
	if (READ_BIT(FLASH->CR, FLASH_CR_LOCK) != 0)
  	{	FLASH->KEYR = FLASH_KEY1;
    	FLASH->KEYR = FLASH_KEY2;
  	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
Lock(void)
{
	WaitForIdle();
	SET_BIT(FLASH->CR, FLASH_CR_LOCK);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void*
FLASH_PageBaseAddress(PageNr page)
{
	return (void*)(FLASH_BASE + (page.number<<CFG_FLASH_PAGE_BITS));
}

//////////////////////////////////////////////////////////////////////////////////////////
PageNr
FLASH_PageNumberOf(const void* address)
{
	PageNr page = { (((UInt32)address-FLASH_BASE)>>CFG_FLASH_PAGE_BITS) };
	return page;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
FLASH_ReadPage(PageNr pageNr, UInt08 data[FLASH_PAGE_SIZE])
{
	const UInt08* flashAddress = FLASH_PageBaseAddress(pageNr);
	WaitForIdle();
	for (UInt32 i=0; i<FLASH_PAGE_SIZE; i++) data[i] = flashAddress[i];
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
FLASH_WritePage(PageNr pageNr, const UInt08 data[FLASH_PAGE_SIZE])
{
	UInt08* flashAddress = FLASH_PageBaseAddress(pageNr);
	// (1) Wenn nötig, Seite vorher löschen
	for (UInt32 i=0; i<FLASH_PAGE_SIZE; i++)
	{   if ((UInt32)(flashAddress[i] && data[i]) != (UInt32)data[i])
		{	FLASH_ErasePage(pageNr);
			break;
		}
	}
	// (2) Seite beschreiben
	Unlock();
	WaitForIdle();
	SET_BIT(FLASH->CR, FLASH_CR_PG);
	UInt16* flashAddress16 = FLASH_PageBaseAddress(pageNr);
	for (UInt32 i=0; i<FLASH_PAGE_SIZE/2; i++)
	{	WaitForIdle();
		UInt16 data16 = ToWord(data[2*i+1], data[2*i]);
		flashAddress16[i] = data16;
	}
    SET_BIT(FLASH->SR, FLASH_SR_EOP);
    CLEAR_BIT(FLASH->CR, FLASH_CR_PG);
	Lock();
	// (3) Geschriebenes verifizieren
	for (UInt32 i=0; i<FLASH_PAGE_SIZE; i++)
	{   if (flashAddress[i] != data[i]) return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
FLASH_ErasePage(PageNr pageNr)	// pageNumber: 0 .. 63
{
	Unlock();
	WaitForIdle();
	SET_BIT(FLASH->CR, FLASH_CR_PER);
    FLASH->AR = (UInt32)FLASH_PageBaseAddress(pageNr);
    SET_BIT(FLASH->CR, FLASH_CR_STRT);
    WaitForIdle();
    SET_BIT(FLASH->SR, FLASH_SR_EOP);
    CLEAR_BIT(FLASH->CR, FLASH_CR_PER);
    Lock();
}

//////////////////////////////////////////////////////////////////////////////////////////
